@@ -4,11 +4,12 @@
  * Declarations relating to support for wide characters; many are simply
  * inherited by (sub-optimal) inclusion of other header files.
  *
- * $Id: wchar.h,v 7f0aa41f8caf 2018/10/21 15:39:35 keith $
+ * $Id: wchar.h,v 03b4e6cd4030 2020/03/12 22:44:03 keith $
  *
  * Unattributed original source.
  * Adapted by Rob Savoye <rob@cygnus.com>
- * Copyright (C) 1997, 1999-2009, 2011, 2015, 2016, 2018, MinGW.org Project.
+ * Copyright (C) 1997, 1999-2009, 2011, 2015, 2016, 2018-2020,
+ *   MinGW.org Project.
  *
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
@@ -528,15 +529,18 @@ typedef wchar_t  _Wint_t;
 
 typedef int mbstate_t;
 
-/* The following multi-byte character conversion functions are
- * implemented in libmingwex.a, (and maybe also in some non-free
- * Microsoft libraries, such as MSVCP60.DLL and later).
+/* The following multi-byte character conversion functions have been
+ * implemented by Microsoft, in non-free MSVCR80.DLL and later, (and
+ * maybe also in some earlier non-free DLLs, such as MSVCP60.DLL and
+ * later); they are also available in MSVCRT.DLL, from Vista onward,
+ * but to provide continuing support for earlier Windows versions,
+ * we invoke them via MinGW specific wrappers, defined below.
  */
 __cdecl __MINGW_NOTHROW  wint_t btowc (int);
 __cdecl __MINGW_NOTHROW  int wctob (wint_t);
 
-__cdecl __MINGW_NOTHROW
-size_t mbrlen (const char *__restrict__, size_t, mbstate_t *__restrict__);
+__cdecl __MINGW_NOTHROW  size_t mbrlen
+(const char *__restrict__, size_t, mbstate_t *__restrict__);
 
 __cdecl __MINGW_NOTHROW  size_t mbrtowc
 (wchar_t *__restrict__, const char *__restrict__, size_t, mbstate_t *__restrict__);
@@ -544,12 +548,115 @@ __cdecl __MINGW_NOTHROW  size_t mbrtowc
 __cdecl __MINGW_NOTHROW  size_t mbsrtowcs
 (wchar_t *__restrict__, const char **__restrict__, size_t, mbstate_t *__restrict__);
 
-__cdecl __MINGW_NOTHROW
-size_t wcrtomb (char * __restrict__, wchar_t, mbstate_t *__restrict__);
+__cdecl __MINGW_NOTHROW  size_t wcrtomb
+(char * __restrict__, wchar_t, mbstate_t *__restrict__);
 
 __cdecl __MINGW_NOTHROW  size_t wcsrtombs
 (char *__restrict__, const wchar_t **__restrict__, size_t, mbstate_t *__restrict__);
 
+/* To provide support for the above, on legacy Windows versions,
+ * we implement fall back wrappers in libmingwex.a; each of these
+ * will delegate to the corresponding Microsoft implementation, if
+ * it exists in the process address space; otherwise, execution
+ * will fall back to a MinGW implementation...
+ */
+__cdecl __MINGW_NOTHROW  wint_t __msvcrt_btowc (int);
+
+__cdecl __MINGW_NOTHROW  size_t __msvcrt_mbrlen
+(const char *__restrict__, size_t, mbstate_t *__restrict__);
+
+__cdecl __MINGW_NOTHROW  size_t __msvcrt_mbrtowc
+(wchar_t *__restrict__, const char *__restrict__, size_t, mbstate_t *__restrict__);
+
+__cdecl __MINGW_NOTHROW  size_t __msvcrt_mbsrtowcs
+(wchar_t *__restrict__, const char **__restrict__, size_t, mbstate_t *__restrict__);
+
+__cdecl __MINGW_NOTHROW  int __msvcrt_wctob (wint_t);
+
+__cdecl __MINGW_NOTHROW  size_t __msvcrt_wcrtomb
+(char * __restrict__, wchar_t, mbstate_t *__restrict__);
+
+__cdecl __MINGW_NOTHROW  size_t __msvcrt_wcsrtombs
+(char *__restrict__, const wchar_t **__restrict__, size_t, mbstate_t *__restrict__);
+
+/* ...whereas, these alternatives will always invoke the MinGW
+ * fall back implementations, without considering any possible
+ * reference to MSVCRT.DLL or MSVCR80.DLL implementations.
+ */
+__cdecl __MINGW_NOTHROW  wint_t __mingw_btowc (int);
+
+__cdecl __MINGW_NOTHROW  size_t __mingw_mbrlen
+(const char *__restrict__, size_t, mbstate_t *__restrict__);
+
+__cdecl __MINGW_NOTHROW  size_t __mingw_mbrtowc
+(wchar_t *__restrict__, const char *__restrict__, size_t, mbstate_t *__restrict__);
+
+__cdecl __MINGW_NOTHROW  size_t __mingw_mbsrtowcs
+(wchar_t *__restrict__, const char **__restrict__, size_t, mbstate_t *__restrict__);
+
+__cdecl __MINGW_NOTHROW  int __mingw_wctob (wint_t);
+
+__cdecl __MINGW_NOTHROW  size_t __mingw_wcrtomb
+(char * __restrict__, wchar_t, mbstate_t *__restrict__);
+
+__cdecl __MINGW_NOTHROW  size_t __mingw_wcsrtombs
+(char *__restrict__, const wchar_t **__restrict__, size_t, mbstate_t *__restrict__);
+
+#if __MSVCRT_VERSION__ < __MSVCR80_DLL
+/* For linking with all versions of MSVCRT.DLL, and with non-free
+ * alternatives predating MSVCR80.DLL, we enforce inline mapping to
+ * the libmingwex.a implementations, (which will delegate the calls
+ * to the Microsoft DLL implementations, when they are available,
+ * but substitute the MinGW replacements as required).
+ */
+#undef __mingw_redirect
+#if _ISOC99_SOURCE & 0x08
+/* The user has explicitly requested ISO-C99 compatibility; ensure
+ * that calls to the following functions are serviced by the MinGW
+ * implementations, regardless of availability of any alternative
+ * MSVCRT.DLL implementation.
+ */
+# define __mingw_redirect(NAME)  __mingw_##NAME
+#else
+/* No explicit ISO-C99 compatibility has been requested; map calls
+ * to these functions to delegate to any MSVCRT.DLL implementations
+ * which may be available, or to fall back to the MinGW replacement
+ * implementations, when necessary.
+ */
+# define __mingw_redirect(NAME)  __msvcrt_##NAME
+#endif
+/* FIXME: Maybe consider these mappings, even for linking with the
+ * non-free MSVCR80.DLL, and its descendants.
+ */
+__CRT_ALIAS __cdecl __MINGW_NOTHROW  wint_t btowc (int __c)
+{ return __mingw_redirect(btowc( __c )); }
+
+__CRT_ALIAS __cdecl __MINGW_NOTHROW  size_t mbrlen
+(const char *__mbc, size_t __n, mbstate_t *__ps)
+{ return __mingw_redirect(mbrlen( __mbc, __n, __ps )); }
+
+__CRT_ALIAS __cdecl __MINGW_NOTHROW  size_t mbrtowc
+(wchar_t *__wc, const char *__mbc, size_t __n, mbstate_t *__ps)
+{ return __mingw_redirect(mbrtowc( __wc, __mbc, __n, __ps )); }
+
+__CRT_ALIAS __cdecl __MINGW_NOTHROW  size_t mbsrtowcs
+(wchar_t *__wcs, const char **__mbs, size_t __n, mbstate_t *__ps)
+{ return __mingw_redirect(mbsrtowcs( __wcs, __mbs, __n, __ps )); }
+
+__CRT_ALIAS __cdecl __MINGW_NOTHROW  int wctob (wint_t __wc)
+{ return __mingw_redirect(wctob( __wc )); }
+
+__CRT_ALIAS __cdecl __MINGW_NOTHROW  size_t wcrtomb
+(char * __mbc, wchar_t __wc, mbstate_t *__ps)
+{ return __mingw_redirect(wcrtomb(__mbc, __wc, __ps)); }
+
+__CRT_ALIAS __cdecl __MINGW_NOTHROW  size_t wcsrtombs
+(char *__mbs, const wchar_t **__wcs, size_t __len, mbstate_t *__ps)
+{ return __mingw_redirect(wcsrtombs(__mbs, __wcs, __len, __ps)); }
+
+#undef	__mingw_redirect
+#endif	/* ! MSVCR80.DLL or later */
+
 #if defined _ISOC99_SOURCE || defined __cplusplus
 /* These ISO-C99 functions are implemented in libmingwex.a,
  * or, in some cases, as inline stubs; while provided as MinGW
@@ -560,13 +667,13 @@ __cdecl __MINGW_NOTHROW  int fwide (FILE *, int);
 __cdecl __MINGW_NOTHROW  int mbsinit (const mbstate_t *);
 
 #ifndef __NO_INLINE__
-__CRT_INLINE __cdecl __MINGW_NOTHROW
-int fwide (FILE *__UNUSED_PARAM(stream), int mode)
-  { return mode; } /* Nothing to do  */
+__CRT_INLINE __LIBIMPL__(( FUNCTION = fwide ))
+__cdecl __MINGW_NOTHROW  int fwide (FILE *__UNUSED_PARAM(__stream), int __mode)
+{ return __mode; } /* Nothing to do  */
 
-__CRT_INLINE __cdecl __MINGW_NOTHROW
-int mbsinit (const mbstate_t *__UNUSED_PARAM(ps))
-  { return 1; }
+__CRT_INLINE __LIBIMPL__(( FUNCTION = mbsinit ))
+__cdecl __MINGW_NOTHROW  int mbsinit (const mbstate_t *__ps)
+{ return ((__ps == NULL) || (*__ps == (mbstate_t)(0))); }
 #endif
 
 __cdecl __MINGW_NOTHROW  wchar_t *wmemset (wchar_t *, wchar_t, size_t);
